#
#    Copyright (C) 2014 Alexandre Viau <alexandre@alexandreviau.net>
#
#    This file is part of python-pass.
#
#    python-pass is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    python-pass is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with python-pass.  If not, see <http://www.gnu.org/licenses/>.
#

import unittest
import os
import shutil
import subprocess
import string
import tempfile

from pypass import PasswordStore
from pypass import EntryType


class TestPasswordStore(unittest.TestCase):

    def setUp(self):
        self.dir = tempfile.mkdtemp()

        # .gpg_id file
        with open(os.path.join(self.dir, '.gpg-id'), 'w') as gpg_id_file:
            gpg_id_file.write('5C5833E3')

        # Create three dummy files
        open(os.path.join(self.dir, 'linux.ca.gpg'), 'a').close()
        open(os.path.join(self.dir, 'passwordstore.org.gpg'), 'a').close()
        open(os.path.join(self.dir, 'test.com.gpg'), 'a').close()

        # Create one folder
        email_folder_path = os.path.join(self.dir, 'Email')
        os.mkdir(email_folder_path)
        open(os.path.join(email_folder_path, 'email.com.gpg'), 'a').close()

    def tearDown(self):
        shutil.rmtree(self.dir)

    def test_constructor(self):
        store = PasswordStore(self.dir)
        self.assertEqual(store.gpg_id, '5C5833E3')
        self.assertFalse(store.uses_git)
        self.assertEqual(self.dir, store.path)

    def test_get_passwords_list(self):
        store = PasswordStore(self.dir)
        self.assertListEqual(
            sorted(store.get_passwords_list()),
            sorted([
                'test.com',
                'linux.ca',
                'passwordstore.org',
                'Email/email.com',
            ])
        )

    def test_encrypt_decrypt(self):
        self.assertFalse(
            os.path.isfile(os.path.join(self.dir, 'hello.com.gpg'))
        )

        store = PasswordStore(self.dir)
        password = 'ELLO'
        store.insert_password('hello.com', password)

        self.assertTrue(
            os.path.isfile(os.path.join(self.dir, 'hello.com.gpg'))
        )

        self.assertEqual(
            password,
            store.get_decypted_password('hello.com')
        )

    def test_get_decypted_password_specific_entry(self):
        store = PasswordStore(self.dir)
        password = 'ELLO'
        store.insert_password('hello.com', password)

        # When there is no 'password:' mention, the password is assumed to be
        # the first line.
        self.assertEqual(
            'ELLO',
            store.get_decypted_password('hello.com', entry=EntryType.password)
        )

        store.insert_password('hello.com', 'sdfsdf\npassword: pwd')
        self.assertEqual(
            'pwd',
            store.get_decypted_password('hello.com', entry=EntryType.password)
        )

        store.insert_password('hello.com', 'sdf\npassword: pwd\nusername: bob')
        self.assertEqual(
            'bob',
            store.get_decypted_password('hello.com', entry=EntryType.username)
        )

    def test_get_decrypted_password_only_password(self):
        store = PasswordStore(self.dir)
        password = 'ELLO'
        store.insert_password('hello.com', password)

    def test_init(self):
        init_dir = tempfile.mkdtemp()
        PasswordStore.init(
            '5C5833E3',
            path=os.path.join(init_dir, '.password-store')
        )

        self.assertTrue(
            os.path.isdir(os.path.join(init_dir, '.password-store'))
        )

        self.assertTrue(
            os.path.isfile(
                os.path.join(init_dir, '.password-store', '.gpg-id')
            )
        )

        self.assertEqual(
            open(
                os.path.join(init_dir, '.password-store', '.gpg-id'),
                'r'
            ).read(),
            '5C5833E3\n'
        )

        shutil.rmtree(init_dir)

    def test_init_clone(self):
        origin_dir = tempfile.mkdtemp()
        destination_dir = tempfile.mkdtemp()

        subprocess.Popen(
            [
                'git',
                '--git-dir=%s' % os.path.join(origin_dir, '.git'),
                '--work-tree=%s' % origin_dir,
                'init',
                origin_dir
            ],
            shell=False
        ).wait()

        open(os.path.join(origin_dir, 'test_git_init_clone.gpg'), 'a').close()

        subprocess.Popen(
            [
                'git',
                '--git-dir=%s' % os.path.join(origin_dir, '.git'),
                '--work-tree=%s' % origin_dir,
                'add', 'test_git_init_clone.gpg',
            ]
        ).wait()

        subprocess.Popen(
            [
                'git',
                '--git-dir=%s' % os.path.join(origin_dir, '.git'),
                '--work-tree=%s' % origin_dir,
                'commit',
                '-m', '"testcommit"',
            ]
        ).wait()

        # Init
        PasswordStore.init(
            path=destination_dir,
            clone_url=origin_dir,
            gpg_id='3CCC3A3A'
        )

        # The key should be imported
        self.assertTrue(
            os.path.isfile(
                os.path.join(destination_dir, 'test_git_init_clone.gpg')
            )
        )

        # The gpg-id file should be created
        self.assertTrue(
            os.path.isfile(
                os.path.join(destination_dir, '.gpg-id')
            )
        )

        shutil.rmtree(origin_dir)
        shutil.rmtree(destination_dir)

    def test_generate_password(self):
        only_letters = PasswordStore.generate_password(
            digits=False,
            symbols=False
        )

        self.assertTrue(only_letters.isalpha())

        alphanum = PasswordStore.generate_password(digits=True, symbols=False)
        self.assertTrue(alphanum.isalnum())
        for char in alphanum:
            self.assertTrue(char not in string.punctuation)

        length_100 = PasswordStore.generate_password(length=100)
        self.assertEqual(len(length_100), 100)



#
#    Copyright (C) 2014 Alexandre Viau <alexandre@alexandreviau.net>
#
#    This file is part of python-pass.
#
#    python-pass is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    python-pass is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with python-pass.  If not, see <http://www.gnu.org/licenses/>.
#

import os
import subprocess
import string
import random
import re

from .entry_type import EntryType

# Find the right gpg binary
if subprocess.call(
        ['which', 'gpg2'],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE) == 0:
    GPG_BIN = 'gpg2'
elif subprocess.call(
        ['which', 'gpg'],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE) == 0:
    GPG_BIN = 'gpg'
else:
    raise Exception("Could not find GPG")


class PasswordStore(object):
    """This is a Password Store

    :param path: The path of the password-store. By default,
                 '$home/.password-store'.
    :param git_dir: The git directory of the password store. By default,
                    it looks for a .git directory in the password store.
    """

    def __init__(
            self,
            path=os.path.join(os.getenv("HOME"), ".password-store"),
            git_dir=None,
    ):
        self.path = path

        # Read the .gpg-id
        gpg_id_file = os.path.join(path, '.gpg-id')
        if os.path.isfile(gpg_id_file):
            self.gpg_id = open(gpg_id_file, 'r').read().strip()
        else:
            raise Exception("could not find .gpg-id file")

        # Try to locate the git dir
        git_dir = git_dir or os.path.join(self.path, '.git')
        self.uses_git = os.path.isdir(git_dir)
        if self.uses_git:
            self.git_dir = git_dir

    def get_passwords_list(self):
        """Returns a list of the passwords in the store

        :returns: Example: ['Email/bob.net', 'example.com']
        """
        passwords = []

        for root, dirnames, filenames in os.walk(self.path):
            for filename in filenames:
                if filename.endswith('.gpg'):
                    path = os.path.join(root, filename.replace('.gpg', ''))
                    simplified_path = path.replace(self.path + '/', '')
                    passwords.append(simplified_path)

        return passwords

    def get_decypted_password(self, path, entry=None):
        """Returns the content of the decrypted password file

        :param path: The path of the password to be decrypted. Example:
                     'email.com'
        :param entry: The entry to retreive. (EntryType enum)
        """
        passfile_path = os.path.realpath(
            os.path.join(
                self.path,
                path + '.gpg'
            )
        )

        gpg = subprocess.Popen(
            [
                GPG_BIN,
                '--quiet',
                '--batch',
                '--use-agent',
                '-d', passfile_path,
            ],
            shell=False,
            stdout=subprocess.PIPE
        )
        gpg.wait()

        if gpg.returncode == 0:
            decrypted_password = gpg.stdout.read().decode()

            if entry == EntryType.username:
                usr = re.search(
                    '(?:username|user|login): (.+)',
                    decrypted_password
                )
                if usr:
                    return usr.groups()[0]
            elif entry == EntryType.password:
                pw = re.search('(?:password|pass): (.+)', decrypted_password)
                if pw:
                    return pw.groups()[0]
                else:  # If there is no match, password is the first line
                    return decrypted_password.split('\n')[0]
            elif entry == EntryType.hostname:
                hostname = re.search(
                    '(?:host|hostname): (.+)', decrypted_password
                )
                return hostname.groups()[0]
            else:
                return decrypted_password

    def insert_password(self, path, password):
        """Encrypts the password at the given path

        :param path: Where to insert the password. Ex: 'passwordstore.org'
        :param password: The password to insert, can be multi-line
        """

        passfile_path = os.path.realpath(
            os.path.join(self.path, path + '.gpg')
        )

        gpg = subprocess.Popen(
            [
                GPG_BIN,
                '-e',
                '-r', self.gpg_id,
                '--batch',
                '--use-agent',
                '--no-tty',
                '--yes',
                '-o', passfile_path
            ],
            shell=False,
            stdin=subprocess.PIPE
        )

        gpg.stdin.write(password.encode())
        gpg.stdin.close()
        gpg.wait()

    @staticmethod
    def generate_password(digits=True, symbols=True, length=15):
        """Returns a random password

        :param digits: Should the password have digits? Defaults to True
        :param symbols: Should the password have symbols? Defaults to True
        :param length: Length of the password. Defaults to 15
        """

        chars = string.ascii_letters

        if symbols:
            chars += string.punctuation

        if digits:
            chars += string.digits

        password = ''.join(random.choice(chars) for i in range(length))
        return password

    @staticmethod
    def init(gpg_id, path, clone_url=None):
        """Creates a password store to the given path

        :param gpg_id: Default gpg key identification used for encryption and
                       decryption. Example: '3CCC3A3A'
        :param path: Where to create the password store. By default, this is
                     $home/.password-store
        :param clone_url: If specified, the clone_url parameter will be used
                          to import a password store from a git repository.
                          Example: ssh://myserver.net:/home/bob/.password-store
        :returns: PasswordStore object
        """
        git_dir = os.path.join(path, '.git')
        git_work_tree = path

        # Create a folder at the path
        if not os.path.exists(path):
            os.makedirs(path)

        # Clone an existing remote repo
        if clone_url:
            # Init git repo
            subprocess.call(
                [
                    "git",
                    "--git-dir=%s" % git_dir,
                    "--work-tree=%s" % git_work_tree,
                    "init", path
                ],
                shell=False
            )

            # Add remote repo
            subprocess.call(
                [
                    "git",
                    "--git-dir=%s" % git_dir,
                    "--work-tree=%s" % git_work_tree,
                    "remote",
                    "add",
                    "origin",
                    clone_url
                ],
                shell=False,
            )

            # Pull remote repo
            # TODO: add parameters for remote and branch ?
            subprocess.call(
                [
                    "git",
                    "--git-dir=%s" % git_dir,
                    "--work-tree=%s" % git_work_tree,
                    "pull",
                    "origin",
                    "master"
                ],
                shell=False
            )

        gpg_id_path = os.path.join(path, '.gpg-id')
        if os.path.exists(gpg_id_path) is False:
            # Create .gpg_id and put the gpg id in it
            with open(os.path.join(path, '.gpg-id'), 'a') as gpg_id_file:
                gpg_id_file.write(gpg_id + '\n')

        return PasswordStore(path)

    def git_init(self, git_dir=None):
        """Transform  the existing password store into a git repository

        :param git_dir: Where to create the git directory. By default, it will
                        be created at the root of the password store in a .git
                        folder.
        """

        self.git_dir = git_dir or os.path.join(self.path, '.git')
        self.uses_git = True

        subprocess.call(
            [
                'git',
                "--git-dir=%s" % self.git_dir,
                "--work-tree=%s" % self.path,
                'init',
            ],
            shell=False
        )

        self.git_add_and_commit(
            '.',
            message="Add current contents of password store."
        )

        # Create .gitattributes and commit it
        with open(
                os.path.join(self.path, '.gitattributes'), 'w'
        ) as gitattributes:
            gitattributes.write('*.gpg diff=gpg\n')

        self.git_add_and_commit(
            '.gitattributes',
            message="Configure git repository for gpg file diff."
        )

        subprocess.call(
            [
                'git',
                "--git-dir=%s" % self.git_dir,
                "--work-tree=%s" % self.path,
                'config',
                '--local',
                'diff.gpg.binary',
                'true'
            ],
            shell=False
        )

        subprocess.call(
            [
                'git',
                "--git-dir=%s" % self.git_dir,
                "--work-tree=%s" % self.path,
                'config',
                '--local',
                'diff.gpg.textconv',
                'gpg -d'
            ],
            shell=False
        )

    def git_add_and_commit(self, path, message=None):

        subprocess.call(
            [
                'git',
                "--git-dir=%s" % self.git_dir,
                "--work-tree=%s" % self.path,
                'add',
                path
            ],
            shell=False
        )

        if message:
            subprocess.call(
                [
                    'git',
                    "--git-dir=%s" % self.git_dir,
                    "--work-tree=%s" % self.path,
                    'commit',
                    '-m',
                    message
                ],
                shell=False
            )
        else:
            subprocess.call(
                [
                    'git',
                    "--git-dir=%s" % self.git_dir,
                    "--work-tree=%s" % self.path,
                    'commit'
                ],
                shell=False
            )



#
#    Copyright (C) 2014 Alexandre Viau <alexandre@alexandreviau.net>
#
#    This file is part of python-pass.
#
#    python-pass is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    python-pass is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with python-pass.  If not, see <http://www.gnu.org/licenses/>.
#

import os
import re
import shutil
import subprocess
import tempfile
import unittest

import click.testing

import pypass.command
import pypass.tests
from pypass.passwordstore import PasswordStore


class TestCommand(unittest.TestCase):

    def run_cli(self, args, input=None):
        args = ['--PASSWORD_STORE_DIR', self.dir] + list(args)
        runner = click.testing.CliRunner()
        result = runner.invoke(pypass.command.main, args, input=input)
        return result

    def setUp(self):
        self.dir = tempfile.mkdtemp()

        # .gpg_id file
        with open(os.path.join(self.dir, '.gpg-id'), 'w') as gpg_id_file:
            gpg_id_file.write('5C5833E3')

    def tearDown(self):
        shutil.rmtree(self.dir)

    def test_init(self):
        init_dir = tempfile.mkdtemp()
        init_result = self.run_cli(
            [
                'init',
                '-p', os.path.join(init_dir, '.password-store'),
                '5C5833E3'
            ]
        )

        self.assertTrue(
            os.path.isdir(os.path.join(init_dir, '.password-store'))
        )

        self.assertTrue(
            os.path.isfile(
                os.path.join(init_dir, '.password-store', '.gpg-id')
            )
        )

        self.assertEqual(
            open(
                os.path.join(init_dir, '.password-store', '.gpg-id'),
                'r'
            ).read(),
            '5C5833E3\n'
        )

        self.assertEqual(
            init_result.output,
            'Password store initialized for 5C5833E3.\n'
        )

        shutil.rmtree(init_dir)

    def test_insert_and_show(self):
        # Insert a password for test.com
        self.run_cli(
            ['insert', 'test.com'],
            input='super_secret\nsuper_secret'
        )

        self.assertTrue(
            os.path.isfile(os.path.join(self.dir, 'test.com.gpg'))
        )

        # Show the password for test.com
        show_result = self.run_cli(
            ['show', 'test.com'],
            input='super_secret\nsuper_secret'
        )

        self.assertEqual(show_result.output, 'super_secret\n')

    def test_show_non_existing_password(self):
        # Show the password for test.com
        show_result = self.run_cli(
            ['show', 'test.com'],
        )

        self.assertEqual(show_result.output,
                         'Error: test.com is not in the password store.\n')

    # Can't get xclip to work in Travis.
    @pypass.tests.skipIfTravis
    def test_show_clip(self):
        store = PasswordStore(self.dir)
        store.insert_password('clip_test', 'clipme999\nbutnotthisnewline\nfff')

        show_result = self.run_cli(['show', '-c', 'clip_test'])

        self.assertEqual(
            show_result.output,
            'Copied clip_test to clipboard.\n'
        )

        # Check if the password is in the clipoard
        xclip = subprocess.Popen(
            ['xclip', '-o', '-selection', 'clipboard'],
            stdout=subprocess.PIPE)
        xclip.wait()
        self.assertEqual(xclip.stdout.read(), 'clipme999')

    def test_edit_not_exist(self):
        edit_result = self.run_cli(
            ['edit', 'woijewoifj.ccc']
        )
        self.assertEqual(
            edit_result.output,
            'woijewoifj.ccc is not in the password store.\n'
        )

    def test_ls(self):
        # Create three dummy files
        open(os.path.join(self.dir, 'linux.ca.gpg'), 'a').close()
        open(os.path.join(self.dir, 'passwordstore.org.gpg'), 'a').close()
        open(os.path.join(self.dir, 'test.com.gpg'), 'a').close()

        ls_result = self.run_cli(['ls'])

        expected_regex = \
            "Password Store\s.*linux.ca\s.*passwordstore.org\s.*test.com"

        self.assertIsNotNone(re.search(expected_regex, ls_result.output))

        # By default, pypass should run the ls command
        ls_default_result = self.run_cli([])
        self.assertEqual(ls_result.output, ls_default_result.output)

    def test_rm(self):
        # Create one dummy file
        dummy_file_path = os.path.join(self.dir, 'test.com.gpg')

        with open(dummy_file_path, 'w') as dummy_file:
            dummy_file.write('test.com')

        self.assertTrue(os.path.isfile(dummy_file_path))
        rm_result = self.run_cli(['rm', 'test.com'], input='y\n')
        self.assertFalse(os.path.isfile(dummy_file_path))
        self.assertIsNotNone(
            re.match("Really remove .*test.com.gpg?", rm_result.output)
        )

    def test_rm_dont_exist(self):
        result = self.run_cli(['rm', 'test.com'])
        self.assertEqual(
            result.output,
            'Error: test.com is not in the password store.\n'
        )

    def test_rm_recursive(self):
        folder_path = os.path.join(self.dir, 'test_folder')
        os.mkdir(folder_path)
        self.assertTrue(os.path.isdir(folder_path))

        # Create three dummy files
        open(os.path.join(folder_path, 'linux.ca.gpg'), 'a').close()
        open(os.path.join(folder_path, 'passwordstore.org.gpg'), 'a').close()
        open(os.path.join(folder_path, 'test.com.gpg'), 'a').close()

        rm_result = self.run_cli(['rm', '-r', 'test_folder'], input='y\n')

        self.assertFalse(os.path.isdir(folder_path))
        self.assertIsNotNone(
            re.match("Recursively remove .*test_folder?", rm_result.output)
        )

    def test_mv_file(self):
        old_file_path = os.path.join(self.dir, 'move_me.gpg')
        open(old_file_path, 'a').close()

        self.assertTrue(os.path.isfile(old_file_path))

        self.run_cli(['mv', 'move_me', 'i_moved'])

        self.assertFalse(os.path.isfile(old_file_path))
        self.assertTrue(os.path.isfile(os.path.join(self.dir, 'i_moved.gpg')))

    def test_mv_folder(self):
        folder_path = os.path.join(self.dir, 'test_folder')
        os.mkdir(folder_path)
        self.assertTrue(os.path.isdir(folder_path))

        # Create three dummy files
        open(os.path.join(folder_path, 'linux.ca.gpg'), 'a').close()
        open(os.path.join(folder_path, 'passwordstore.org.gpg'), 'a').close()
        open(os.path.join(folder_path, 'test.com.gpg'), 'a').close()

        self.run_cli(['mv', 'test_folder', 'moved_folder'])

        self.assertFalse(os.path.isdir(folder_path))
        self.assertTrue(os.path.isdir(os.path.join(self.dir, 'moved_folder')))

    def test_mv_error(self):
        mv_result = self.run_cli(['mv', 'test_folder', 'moved_folder'])
        self.assertEqual(
            mv_result.output,
            'Error: test_folder is not in the password store\n'
        )

    def test_cp_file(self):
        old_file_path = os.path.join(self.dir, 'copy_me.gpg')
        open(old_file_path, 'a').close()

        self.assertTrue(os.path.isfile(old_file_path))

        self.run_cli(['cp', 'copy_me', 'i_was_copied'])

        self.assertTrue(os.path.isfile(old_file_path))
        self.assertTrue(
            os.path.isfile(os.path.join(self.dir, 'i_was_copied.gpg'))
        )

    def test_cp_folder(self):
        folder_path = os.path.join(self.dir, 'test_folder')
        os.mkdir(folder_path)
        self.assertTrue(os.path.isdir(folder_path))

        # Create three dummy files
        open(os.path.join(folder_path, 'linux.ca.gpg'), 'a').close()
        open(os.path.join(folder_path, 'passwordstore.org.gpg'), 'a').close()
        open(os.path.join(folder_path, 'test.com.gpg'), 'a').close()

        self.run_cli(['cp', 'test_folder', 'copied_folder'])

        self.assertTrue(os.path.isdir(folder_path))
        self.assertTrue(os.path.isdir(os.path.join(self.dir, 'copied_folder')))

    def test_cp_error(self):
        mv_result = self.run_cli(['cp', 'test_folder', 'moved_folder'])
        self.assertEqual(
            mv_result.output,
            'Error: test_folder is not in the password store\n'
        )

    def test_find(self):
        # Create dummy files
        open(os.path.join(self.dir, 'linux.ca.gpg'), 'a').close()
        open(os.path.join(self.dir, 'passwordstore.org.gpg'), 'a').close()
        open(os.path.join(self.dir, 'test.com.gpg'), 'a').close()
        open(os.path.join(self.dir, 'vv.com.gpg'), 'a').close()
        open(os.path.join(self.dir, 'zz.com.gpg'), 'a').close()

        find_result = self.run_cli(['find', 'pass', 'vv'])

        expected_regex = \
            "Search\sTerms:\spass,vv\s.*passwordstore.org\s.*vv.com"

        self.assertIsNotNone(re.search(expected_regex, find_result.output))

    def test_grep(self):
        store = PasswordStore(self.dir)
        store.insert_password('grep_test.com', 'GREPME')

        grep_result = self.run_cli(['grep', 'GREPME'])
        self.assertEqual(
            grep_result.output,
            'grep_test.com:\nGREPME\n'
        )

    def test_git_init(self):
        self.run_cli(['git', 'init'])

        # git init should add a .gitattributes file
        self.assertEqual(
            open(os.path.join(self.dir, '.gitattributes'), 'r').read(),
            '*.gpg diff=gpg\n'
        )

        # git init should set diff.gpg.binary to True
        diff_gpg_binary = subprocess.Popen(
            [
                'git',
                '--git-dir=%s' % os.path.join(self.dir, '.git'),
                '--work-tree=%s' % self.dir,
                'config',
                '--local',
                'diff.gpg.binary'
            ],
            shell=False,
            stdout=subprocess.PIPE
        )
        diff_gpg_binary.wait()
        self.assertEqual(diff_gpg_binary.stdout.read().decode(), 'true\n')

        # git init should set diff.gpg.textconv to 'gpg -d'
        gpg = subprocess.Popen(
            [
                'git',
                '--git-dir=%s' % os.path.join(self.dir, '.git'),
                '--work-tree=%s' % self.dir,
                'config',
                '--local',
                'diff.gpg.textconv'
            ],
            shell=False,
            stdout=subprocess.PIPE
        )
        gpg.wait()
        self.assertEqual(gpg.stdout.read().decode(), 'gpg -d\n')

    def test_init_clone(self):
        # Setup origin repo
        origin_dir = tempfile.mkdtemp()
        origin_git_dir = os.path.join(origin_dir, '.git')

        subprocess.Popen(
            [
                'git',
                '--git-dir=%s' % origin_git_dir,
                '--work-tree=%s' % origin_dir,
                'init',
                origin_dir
            ],
            shell=False
        ).wait()

        open(os.path.join(origin_dir, 'test_git_init_clone.gpg'), 'a').close()

        subprocess.call(
            [
                'git',
                '--git-dir=%s' % origin_git_dir,
                '--work-tree=%s' % origin_dir,
                'add', 'test_git_init_clone.gpg',
            ]
        )

        subprocess.call(
            [
                'git',
                '--git-dir=%s' % origin_git_dir,
                '--work-tree=%s' % origin_dir,
                'commit',
                '-m', '"testcommit"',
            ]
        )

        # Init
        self.run_cli(
            [
                'init',
                '--path', self.dir,
                '--clone', origin_dir,
                'TEST_GPG_ID'
            ]
        )

        # The key should be imported
        self.assertTrue(
            os.path.isfile(
                os.path.join(self.dir, 'test_git_init_clone.gpg')
            )
        )

        # The gpg-id file should be created
        self.assertTrue(
            os.path.isfile(
                os.path.join(self.dir, '.gpg-id')
            )
        )



#
#    Copyright (C) 2014 Alexandre Viau <alexandre@alexandreviau.net>
#
#    This file is part of python-pass.
#
#    python-pass is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    python-pass is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with python-pass.  If not, see <http://www.gnu.org/licenses/>.
#

import os
import unittest

skipIfTravis = unittest.skipIf(
    os.getenv('TRAVIS') == 'true', "Skipping this test on travis."
)



#
#    Copyright (C) 2014 Alexandre Viau <alexandre@alexandreviau.net>
#
#    This file is part of python-pass.
#
#    python-pass is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    python-pass is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with python-pass.  If not, see <http://www.gnu.org/licenses/>.
#

import os
import subprocess
import shutil
import sys
import tempfile

import click
import colorama
import pxssh

from pypass.entry_type import EntryType
from pypass import PasswordStore


@click.group(invoke_without_command=True)
@click.option('--PASSWORD_STORE_DIR',
              envvar='PASSWORD_STORE_DIR',
              default=os.path.join(os.getenv("HOME"), ".password-store"),
              type=click.Path(file_okay=False, resolve_path=True))
@click.option('--PASSWORD_STORE_GIT',
              envvar='PASSWORD_STORE_GIT',
              type=click.Path(file_okay=False, resolve_path=True),
              default=None)
@click.option('--EDITOR',
              envvar='EDITOR',
              default='editor',
              type=click.STRING)
@click.pass_context
def main(ctx, password_store_dir, password_store_git, editor):

    # init does not need any of this.
    if ctx.invoked_subcommand == "init":
        return

    # Prepare the config file
    config = {
        'password_store': PasswordStore(
            path=password_store_dir,
            git_dir=password_store_git
        ),
        'editor': editor
    }

    ctx.obj = config

    # By default, invoke ls
    if ctx.invoked_subcommand is None:
        ctx.invoke(ls)


@main.command(name='help')
@click.pass_context
def hlp(contex):
    click.echo(contex.parent.get_help())


@main.command()
@click.option('--path', '-p',
              type=click.Path(file_okay=False, resolve_path=True),
              default=os.path.join(os.getenv("HOME"), ".password-store"),
              help='Where to create the password store.')
@click.option('--clone', '-c',
              type=click.STRING,
              help='Git url to clone')
@click.argument('gpg-id', type=click.STRING)
def init(path, clone, gpg_id):
    PasswordStore.init(gpg_id, path, clone_url=clone)
    click.echo("Password store initialized for %s." % gpg_id)


@main.command()
@click.option('--multiline', '-m', is_flag=True)
@click.argument('path', type=click.STRING)
@click.pass_obj
def insert(config, path, multiline):

    if multiline:
        with tempfile.NamedTemporaryFile() as temp_file:
            if subprocess.call([config['editor'], temp_file.name]) == 0:
                password = temp_file.read().strip()
    else:
        password = click.prompt(
            'Enter the password',
            type=str,
            confirmation_prompt=True,
            hide_input=True
        )

    config['password_store'].insert_password(path, password)

    if config['password_store'].uses_git:
        config['password_store'].git_add_and_commit(
            path + '.gpg',
            message='Added %s to store' % path
        )


@main.command()
@click.option('--no-symbols', '-n', is_flag=True, default=False)
@click.argument('pass_name', type=click.STRING)
@click.argument('pass_length', type=int)
def generate(pass_name, pass_length, no_symbols):
    symbols = not no_symbols

    password = PasswordStore.generate_password(
        digits=True,
        symbols=symbols,
        length=pass_length
    )

    print (password)


@main.command()
@click.pass_obj
@click.argument('path', type=click.STRING)
def edit(config, path):
    if path in config['password_store'].get_passwords_list():
        old_password = config['password_store'].get_decypted_password(path)
        with tempfile.NamedTemporaryFile() as temp_file:
            temp_file.write(old_password)
            temp_file.flush()

            subprocess.call([config['editor'], temp_file.name])
            temp_file.seek(0)

            config['password_store'].insert_password(
                path, temp_file.file.read()
            )
            click.echo("%s was updated." % path)

            if config['password_store'].uses_git:
                config['password_store'].git_add_and_commit(
                    path + '.gpg',
                    message='Edited password for %s using %s'
                            % (path, config['editor'])
                )
    else:
        click.echo("%s is not in the password store." % path)


@main.command()
@click.option('--clip', '-c', is_flag=True)
@click.argument('path', type=click.STRING)
@click.pass_obj
def show(config, path, clip):
    if path not in config['password_store'].get_passwords_list():
        click.echo('Error: %s is not in the password store.' % path)
        sys.exit()

    decrypted_password = \
        config['password_store'].get_decypted_password(path).strip()

    if clip:
        xclip = subprocess.Popen(
            [
                'xclip',
                '-selection', 'clipboard'
            ],
            stdin=subprocess.PIPE
        )
        xclip.stdin.write(decrypted_password.split('\n')[0])
        xclip.stdin.close()
        click.echo('Copied %s to clipboard.' % path)
    else:
        click.echo(decrypted_password)


@main.command()
@click.argument('path', type=click.STRING)
@click.pass_obj
def connect(config, path):
    store = config['password_store']
    hostname = store.get_decypted_password(path, entry=EntryType.hostname)
    username = store.get_decypted_password(path, entry=EntryType.username)
    password = store.get_decypted_password(path, entry=EntryType.password)
    s = pxssh.pxssh()
    click.echo("Connectig to %s" % hostname)
    s.login(hostname, username, password=password)
    s.sendline()
    s.interact()


@main.command()
@click.argument('subfolder', required=False, type=click.STRING, default='')
@click.pass_obj
def ls(config, subfolder):
    tree = subprocess.Popen(
        [
            'tree',
            '-C',
            '-l',
            '--noreport',
            os.path.join(config['password_store'].path, subfolder),
        ],
        shell=False,
        stdout=subprocess.PIPE
    )
    tree.wait()

    if tree.returncode == 0:
        output_without_gpg = \
            tree.stdout.read().decode('utf8').replace('.gpg', '')

        output_replaced_first_line =\
            "Password Store\n" + '\n'.join(output_without_gpg.split('\n')[1:])

        output_stripped = output_replaced_first_line.strip()

        click.echo(output_stripped)


@main.command()
@click.argument('search_terms', nargs=-1)
@click.pass_obj
def find(config, search_terms):
    click.echo("Search Terms: " + ','.join(search_terms))

    pattern = '*' + '*|*'.join(search_terms) + '*'

    tree = subprocess.Popen(
        [
            'tree',
            '-C',
            '-l',
            '--noreport',
            '-P', pattern,
            # '--prune', (tree>=1.5)
            # '--matchdirs', (tree>=1.7)
            # '--ignore-case', (tree>=1.7)
            config['password_store'].path,
        ],
        shell=False,
        stdout=subprocess.PIPE
    )
    tree.wait()

    if tree.returncode == 0:
        output_without_gpg = \
            tree.stdout.read().decode('utf8').replace('.gpg', '')

        output_without_first_line =\
            '\n'.join(output_without_gpg.split('\n')[1:]).strip()

        click.echo(output_without_first_line)


@main.command()
@click.argument('search_string')
@click.pass_obj
def grep(config, search_string):
    for password in config['password_store'].get_passwords_list():
        decrypted_password = \
            config['password_store'].get_decypted_password(password)

        grep = subprocess.Popen(
            ['grep', '-e', search_string],
            stdout=subprocess.PIPE,
            stdin=subprocess.PIPE
        )

        grep.stdin.write(decrypted_password.encode())
        grep.stdin.close()

        grep_stdout = grep.stdout.read().decode().strip()
        if grep_stdout.strip() != '':
            click.echo(
                colorama.Fore.BLUE + password + ":" + '\n' +
                colorama.Fore.RESET + grep_stdout
            )


@main.command()
@click.option('--recursive', '-r', is_flag=True)
@click.argument('path', type=click.STRING)
@click.pass_obj
def rm(config, recursive, path):
    resolved_path = os.path.realpath(
        os.path.join(config['password_store'].path, path)
    )

    if os.path.isdir(resolved_path) is False:
        resolved_path = os.path.join(
            config['password_store'].path,
            path + '.gpg'
        )

    if os.path.exists(resolved_path):
        if recursive:
            click.confirm("Recursively remove %s?" % resolved_path, abort=True)
            shutil.rmtree(resolved_path)
        else:
            click.confirm("Really remove %s?" % resolved_path, abort=True)
            os.remove(resolved_path)
        click.echo("%s was removed from the store." % path)
    else:
        click.echo("Error: %s is not in the password store." % path)


@main.command()
@click.argument('old_path', type=click.STRING)
@click.argument('new_path', type=click.STRING)
@click.pass_obj
def cp(config, old_path, new_path):
    resolved_old_path = os.path.realpath(
        os.path.join(config['password_store'].path, old_path)
    )

    if os.path.isdir(resolved_old_path):
        shutil.copytree(
            resolved_old_path,
            os.path.realpath(
                os.path.join(config['password_store'].path, new_path)
            )
        )
    else:
        resolved_old_path = os.path.realpath(
            os.path.join(config['password_store'].path, old_path + '.gpg')
        )

        if os.path.isfile(resolved_old_path):
            shutil.copy(
                resolved_old_path,
                os.path.realpath(
                    os.path.join(
                        config['password_store'].path,
                        new_path + '.gpg'
                    )
                )
            )
        else:
            click.echo("Error: %s is not in the password store" % old_path)


@main.command()
@click.argument('old_path', type=click.STRING)
@click.argument('new_path', type=click.STRING)
@click.pass_obj
def mv(config, old_path, new_path):
    resolved_old_path = os.path.realpath(
        os.path.join(config['password_store'].path, old_path)
    )

    if os.path.isdir(resolved_old_path):
        shutil.move(
            resolved_old_path,
            os.path.realpath(
                os.path.join(config['password_store'].path, new_path)
            )
        )
    else:
        resolved_old_path = os.path.realpath(
            os.path.join(config['password_store'].path, old_path + '.gpg')
        )

        if os.path.isfile(resolved_old_path):
            shutil.move(
                resolved_old_path,
                os.path.realpath(
                    os.path.join(
                        config['password_store'].path,
                        new_path + '.gpg'
                    )
                )
            )
        else:
            click.echo("Error: %s is not in the password store" % old_path)


@main.command()
@click.argument('commands', nargs=-1)
@click.pass_obj
def git(config, commands):
    command_list = list(commands)

    if len(command_list) > 0 and command_list[0] == 'init':
        config['password_store'].git_init()
    else:
        subprocess.call(
            [
                'git',
                '--git-dir=%s' % config['password_store'].git_dir,
                '--work-tree=%s' % config['password_store'].path,
            ] + command_list,
            shell=False,
        )


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-

import os
import sys

import sphinx_rtd_theme

sys.path.insert(0, os.path.abspath('../..'))
# -- General configuration ----------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom ones.
extensions = ['sphinx.ext.autodoc']


# autodoc generation is a bit aggressive and a nuisance when doing heavy
# text edit cycles.
# execute "export SPHINX_DEBUG=1" in your terminal to disable

# The suffix of source filenames.
source_suffix = '.rst'

# The master toctree document.
master_doc = 'index'

# General information about the project.
project = u'pypass'
# copyright = u'2014, Alexandre Viau'

# If true, '()' will be appended to :func: etc. cross-reference text.
add_function_parentheses = True

# If true, the current module name will be prepended to all description
# unit titles (such as .. function::).
add_module_names = True

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'

# -- Options for HTML output --------------------------------------------------

# The theme to use for HTML and HTML Help pages.  Major themes that come with
# Sphinx are currently 'default' and 'sphinxdoc'.
html_theme_path = [sphinx_rtd_theme.get_html_theme_path()]
html_theme = 'sphinx_rtd_theme'
html_static_path = ['_static']

# Output file base name for HTML help builder.
htmlhelp_basename = '%sdoc' % project

# -- Options for manual page output -------------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
    (
        'manpage',
        'pypass',
        u'stores, retrieves, generates, and synchronizes passwords securely',
        [],
        1
    )
]

# Example configuration for intersphinx: refer to the Python standard library.
intersphinx_mapping = {'http://docs.python.org/': None}
